/**
 * POP3Session - Class for checking e-mail via POP3 protocol.

	Written by D. Kovachev
 */

import java.io.*;
import java.net.*;
import java.util.*;
import PingThread;

public class POP3Session extends Object
{
    /** 15 sec. socket read timeout */
    public static final int SOCKET_READ_TIMEOUT = 15*1000;

    protected Socket pop3Socket;
    protected BufferedReader in;
    protected PrintWriter out;

    private String host;
    private int port;
    private String userName;
    private String password;


    /**
     * Creates new POP3 session by given POP3 host, username and password.
     * Assumes POP3 port is 110 (default for POP3 service).
     */
    public POP3Session(String host, String userName, String password)
    {
        this(host, 110, userName, password);
    }

    /**
     * Creates new POP3 session by given POP3 host and port, username and password.
     */
    public POP3Session(String host, int port, String userName, String password)
    {
        this.host = host;
        this.port = port;
        this.userName = userName;
        this.password = password;
    }

    /**
     * Throws exception if given server response if negative. According to POP3
     * protocol, positive responses start with a '+' and negative start with '-'.
     */
    protected void checkForError(String response)
    throws IOException
    {
		try
		{
			if (response.charAt(0) != '+')
				throw new IOException(response);
		}
		catch (Exception e)
		{
			throw new IOException("No response");
		}
    }
	
	/**
     * Get headers returns a list of message numbers along with some sizing
     * information, and possibly other information depending on the server.
	 */
    public String[] getHeaders()
    throws IOException
    {
		synchronized (pop3Socket)
		{
			doCommand("LIST");
			return getMultilineResponse();
		}
	}
	
    /**
     * Retrieves the entire text of a message using the POP3 RETR command.
     */
    public String getMessage(String messageId)
    throws IOException
    {
		synchronized (pop3Socket)
		{
			doCommand("RETR " + messageId);
			String[] messageLines = getMultilineResponse();
			StringBuffer message = new StringBuffer();
			for (int i=0; i<messageLines.length; i++) {
				message.append(messageLines[i]);
				message.append("\n");
			}
			return new String(message);
		}
    }

    /**
     * Deletes a particular message with DELE command.
     */
    public void deleteMessage(String messageId)
    throws IOException
    {
		synchronized (pop3Socket)
		{
			doCommand("DELE " + messageId);
		}
    }

    /**
     * Initiates a graceful exit by sending QUIT command.
     */
    public void quit()
    throws IOException
    {
		synchronized (pop3Socket)
		{
			doCommand("QUIT");
		}
    }

    /**
     * Connects to the POP3 server and logs on it
     * with the USER and PASS commands.
     */
    public void connectAndAuthenticate()
    throws IOException
    {
        // Make the connection (to the proxy)
		pop3Socket = new Socket("localhost", 3004);
        pop3Socket.setSoTimeout(SOCKET_READ_TIMEOUT);
        in = new BufferedReader(new InputStreamReader(pop3Socket.getInputStream()));
        out = new PrintWriter(new OutputStreamWriter(pop3Socket.getOutputStream()));
		
		// Connect through the proxy
        out.println(host + ":" + port);
        out.flush();
		
        // Receive the welcome message
        String response = in.readLine();
        checkForError(response);

        // Send a USER command to authenticate
        doCommand("USER " + userName);

        // Send a PASS command to finish authentication
		doCommand("PASS " + password);
		
		(new PingThread(pop3Socket)).start();
    }

    /**
     * Closes down the connection to POP3 server (if open).
     * Should be called if an exception is raised during the POP3 session.
     */
    public void close()
    {
        try {
            in.close();
            out.close();
            pop3Socket.close();
        } catch (Exception ex) {
            // Ignore the exception. Probably the socket is not open.
        }
    }

    /**
     * Sends a POP3 command and retrieves the response. If the response is
     * negative (begins with '-'), throws an IOException with received response.
     */
    protected String doCommand(String command)
    throws IOException
    {
        out.println(command);
        out.flush();
        String response = in.readLine();
        checkForError(response);
        return response;
    }

    /**
     * Retrieves a multi-line POP3 response. If a line contains "." by itself,
     * it is the end of the response. If a line starts with a ".", it should
     * really have two "."'s. We strip off the leading ".". If a line does not
     * start with ".", there should be at least one line more.
     */
    protected String[] getMultilineResponse()
    throws IOException
    {
		Vector allLines = new Vector();

        while (true) {
            String line = in.readLine();

            if (line == null) {
                // Server closed connection
                throw new IOException("Server unawares closed the connection.");
            }
            if (line.equals(".")) {
                // No more lines in the server response
                break;
            }
            if ((line.length() > 0) && (line.charAt(0) == '.')) {
                // The line starts with a "." - strip it off.
                line = line.substring(1);
            }

            // Add read line to the list of lines
			allLines.addElement(line);
        }
		
		String response[] = new String[allLines.size()];
		for (int i = 0; i < allLines.size(); i++)
			response[i] = (String) allLines.elementAt(i);

        return response;
    }

}


