package nakov;

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

/**
 * NakovTCPProxyServer is TCP proxy  server designed to provide forward services
 * to its clients. When a client is connected, it should send a single line request
 * for connection to some remote server in format "server:port" or "IP:port".
 * After that NakovTCPProxyServer establishes a connection to the requested server
 * and starts forwarding the traffic between the client and this server. All data
 * received from the server is sent to the client and all data received from the
 * server is sent to the client. So the client can consider that it is connected
 * directly to the server and no TCP proxy server is present. NakovTCPProxyServer
 * can be used to pass over a firewall or some security permissions. For example
 * a Java applet running in a web browser can use this server to open TCP connections
 * to the Internet and for example to access POP3 mailboxes on remote server.
 */

public class NakovTCPProxyServer {
  public static final int SERVER_PORT = 3004;

  public static void main(String[] aArgs) throws Exception {
    ServerSocket serverSocket = new ServerSocket(SERVER_PORT);
    System.out.println("Nakov TCP Proxy Server is running on port " + SERVER_PORT);
    while (true) {
      Socket socket = serverSocket.accept();
      ClientThread clientThread = new ClientThread(socket);
      clientThread.start();
    }
  }
}


/**
 * Handles client connections. Reads a pair server:port (as single text line)
 * from the client's socket, connects to the specified server and port and does
 * data forwarding between the client's socket and specified server in both
 * directions ("client in <--> server out" and "server in <--> client out")
 * until one of the parties closes the socket or I/O error occurs.
 */
class ClientThread extends Thread {

  private Socket mClientSocket = null;
  private Socket mServerSocket = null;
  private String mClientHostPort;
  private String mServerHostPort;
  private boolean mBothConnectionsAreAlive;

  public ClientThread(Socket aClientSocket) {
    mClientSocket = aClientSocket;
  }

  /**
   * Obtains a destination server socket to some of the servers in the list
   * and starts two threads for forwarding : "client in <--> dest server out"
   * and "dest server in <--> client out". After that this thread dies.
   */
  public void run() {
    try {
      mClientHostPort = mClientSocket.getInetAddress().getHostAddress() + ":" + mClientSocket.getPort();
      // Open input and output streams for the client connection
      BinaryTextInputStream clientIn = new BinaryTextInputStream(mClientSocket.getInputStream());
      OutputStream clientOut = mClientSocket.getOutputStream();
      // Read a server and port from the client socket in format IP:port
      mServerHostPort = clientIn.readLine();
      // Create a new socket connection to the server specified by the client
      mServerSocket = connectToServer();
      if (mServerSocket == null) {  // If connection can not be established
        System.out.println("Can not establish connection to " +
                           mServerHostPort + " for client " + mClientHostPort + ".");
        try {
          mClientSocket.close();
        } catch (IOException e) {
        }
        return;
      }
      // Open input and output streams for the server connection
      BinaryTextInputStream serverIn =
              new BinaryTextInputStream(mServerSocket.getInputStream());
      OutputStream serverOut = mServerSocket.getOutputStream();
      // Start forwarding of socket data between server and client
      ForwardThread clientForward =
              new ForwardThread(this, clientIn, serverOut);
      ForwardThread serverForward =
              new ForwardThread(this, serverIn, clientOut);
      mBothConnectionsAreAlive = true;
      clientForward.start();
      serverForward.start();
      System.out.println("TCP forwarding  " + mClientHostPort +
                         " <--> " + mServerHostPort + "  started.");
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }


  /**
   * connectionBroken() method is called by forwarding child threads to notify
   * this thread (their parent thread) that one of the connections (server or
   * client) is broken (a read/write failure occured). This method disconnects
   * both server and client sockets causing both threads to stop forwarding.
   */
  public synchronized void connectionBroken() {
    try {
      mServerSocket.close();
    } catch (IOException e) {
    }
    try {
      mClientSocket.close();
    } catch (IOException e) {
    }
    if (mBothConnectionsAreAlive) {
      System.out.println("TCP forwarding  " + mClientHostPort +
                         " <--> " + mServerHostPort + "  stopped.");
      mBothConnectionsAreAlive = false;
    }
  }

  /**
   * Reads a server and port from the client socket in format IP:port,
   * connects to the specified by the client server and returns the socket
   * connection. Returns null if a connection can not be established.
   */
  private Socket connectToServer() {
    Socket socket;
    try {
      StringTokenizer stServerPort = new StringTokenizer(mServerHostPort, ": ");
      String host = stServerPort.nextToken();
      int port = Integer.parseInt(stServerPort.nextToken());
      socket = new Socket(host, port);
    } catch (Exception ex) {
      socket = null;
    }
    return socket;
  }
}

/**
 * BinaryTextInputStream is a mixed binary/text input stream.
 * It has a binary read(byte[]) method and a text readLine() method.
 */
class BinaryTextInputStream {
  protected InputStream mIn;

  public BinaryTextInputStream(InputStream aIn) {
    mIn = aIn;
  }

  /**
   * Reads a text line from the stream.
   * @return the line read or null if the end of stream is reached.
   * Considers the line finishes with either "\n" or "\r\n".
   * @throw IOException in case of I/O problem.
   */

  public String readLine() throws IOException {
    StringBuffer line = new StringBuffer();
    while (true) {
      int nextByte = mIn.read();
      if (nextByte == -1) // EOF
        if (line.length() == 0)
          return null;
        else
          break;
      char ch = (char) nextByte;
      if (ch == '\n')
        break;
      line.append(ch);
    }
    if ((line.length() > 0) && (line.charAt(line.length() - 1) == '\r'))
      line.setLength(line.length() - 1);
    return new String(line);
  }

  /**
   * Delegates to InputStream.read(byte[] b).
   */
  public int read(byte[] buf) throws IOException {
    return mIn.read(buf);
  }
}

/**
 * ForwardThread handles the TCP forwarding between a socket input stream (source)
 * and a socket output stream (destination). It reads the input stream and forwards
 * everything to the output stream. If some of the streams fails, the forwarding
 * is stopped and the parent thread is notified to close all its connections.
 */
class ForwardThread extends Thread {
  private static final int READ_BUFFER_SIZE = 8192;

  private BinaryTextInputStream mInputStream = null;
  private OutputStream mOutputStream = null;
  private ClientThread mParent = null;

  /**
   * Creates a new traffic forward thread specifying its input stream,
   * output stream and parent thread
   */
  public ForwardThread(ClientThread aParent, BinaryTextInputStream aInputStream,
                       OutputStream aOutputStream) {
    mInputStream = aInputStream;
    mOutputStream = aOutputStream;
    mParent = aParent;
  }

  /**
   * Runs the thread. Until it is possible, reads the input stream and puts read
   * data in the output stream. If reading can not be done (due to exception or
   * when the stream is at his end) or writing is failed, exits the thread.
   */
  public void run() {
    byte[] buffer = new byte[READ_BUFFER_SIZE];
    try {
      while (true) {
        int bytesRead = mInputStream.read(buffer);
        if (bytesRead == -1)
          break; // Connection is broken. Exit the thread
        mOutputStream.write(buffer, 0, bytesRead);
      }
    } catch (IOException e) {
      // Read/write failed --> connection is broken --> exit the thread
    }
    // Notify the parent thread that the connection is broken
    // and the forwarding should stop
    mParent.connectionBroken();
  }
}
