package nick.inetjava.pop3applet;

import java.net.*;
import java.io.*;
import java.util.*;

/**
 * Takes POP3Command-s from the queue and executes them.
 *
 * @author Nick Nickolov, nick_nickolov@abv.bg
 * @version May 30, 2002 8:24:19 PM
 */
final class POP3Thread extends Thread {
  private static final int PORT = 110;
  private static final String[] STRING_ARRAY_0 = new String[0];
  private POP3Applet owner;
  private Socket socket;
  private String host;
  private String user;
  private String password;
  private BufferedReader in;
  private PrintStream out;
  private Vector v = new Vector();
  private Queue queue;

  POP3Thread(POP3Applet owner, Queue queue) {
    this.owner = owner;
    this.queue = queue;
  }

  public void run() {
    while (true) {
      POP3Command command = (POP3Command) queue.pop();
      if (POP3Applet.DEBUG) {
        System.out.println("DUMP executing " + command.getCommand());
      }
      try {
        Object result = null;
        switch (command.getCommand()) {
          case POP3Command.C_CHECK_MAIL: {
            Object[] arguments = command.getArguments();
            result = onCheckMail((String) arguments[0], (String) arguments[1], (String) arguments[2]);
            break;
          }
          case POP3Command.C_DISCONNECT: {
            onDisconnect();
            break;
          }
          case POP3Command.C_DELETE_MESSAGE: {
            onDeleteMessage(((Integer) command.getArguments()[0]).intValue());
            break;
          }
          case POP3Command.C_SHOW_MESSAGE: {
            result = onShowMessage(((Integer) command.getArguments()[0]).intValue());
            break;
          }
          case POP3Command.C_KEEP_CONNECTION: {
            onKeepConnection();
            break;
          }
          default: {
            break;
          }
        }
        owner.commandCompleted(command, true, result);
      } catch (IOException e) {
        if (POP3Applet.DEBUG) {
          e.printStackTrace();
        }
        String s = e.getMessage();
        if (s == null) {
          s = e.getClass().getName();
        }
        owner.commandCompleted(command, false, s);
      }
    }
  }

  /**
   * Connects and authorizes only if necessary.
   */
  private String[] onCheckMail(String host, String user, String password) throws IOException {
    if ((host == null) || (user == null) || (password == null)) {
      throw new IllegalArgumentException();
    }
    boolean needsConnecting = (socket == null) || !host.equals(this.host);
    boolean needsAuthorization = needsConnecting || !user.equals(this.user) || !password.equals(this.password);
    if (needsConnecting) {
      owner.setStatus("Connecting to " + host + ":" + PORT + " ...");
      this.host = host;
      socket = (POP3Applet.USE_NAKOV_TCP_PROXY)
                        ? new Socket(POP3Applet.NAKOV_TCP_PROXY_HOST,
                          POP3Applet.NAKOV_TCP_PROXY_PORT)
                        : new Socket(host, PORT);
      in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
      out = new PrintStream(socket.getOutputStream());
      if (POP3Applet.USE_NAKOV_TCP_PROXY) {
        out.println(host + ':' + PORT);
      }
    }
    if (needsAuthorization) {
      owner.setStatus("Authorizing as '" + user + "' ...");
      this.user = user;
      this.password = password;
      readReply(); // skip greeting line
      doCommand("user " + user);
      doCommand("pass " + password);
    }
    String[] a = doMultiLineCommand("list");
    owner.setStatus("Mail checked.");
    return a;
  }

  private void onDisconnect() throws IOException {
    owner.setStatus("Disconnecting...");
    if (socket != null) {
      try {
        doCommand("quit");
      } catch (IOException ioe) {
        // ignore
      }
      try {
        socket.close();
        in.close();
        out.close();
      } catch (IOException ioe) {
        // ignore
      }
    }
    socket = null;
    in = null;
    out = null;
    owner.setStatus("Disconnected.");
  }

  /**
   * @param id 0-based
   */
  private void onDeleteMessage(int id) throws IOException {
    id++;
    owner.setStatus("Deleting message " + id + " ...");
    doCommand("dele " + id);
    owner.setStatus("Message " + id + " deleted.");
  }

  /**
   * @param id 0-based
   */
  private String onShowMessage(int id) throws IOException {
    id++;
    owner.setStatus("Retrieving message " + id + " ...");
    String[] a = doMultiLineCommand("retr " + id);
    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < a.length; i++) {
      buffer.append(a[i]).append('\n');
    }
    owner.setStatus("Message " + id + " retrieved.");
    return buffer.toString();
  }

  private void onKeepConnection() throws IOException {
    if (POP3Applet.DEBUG) {
      System.out.println("DUMP keep connection");
    }
    if (socket != null) {
      doCommand("stat");
    }
  }

  private String[] doMultiLineCommand(String s) throws IOException {
    String r = doCommand(s);
    v.setSize(0);
    while (true) {
      r = in.readLine();
      if (in == null) {
        throw new POP3Exception();
      }
      if (r.equals(".")) {
        break;
      }
      v.addElement(r);
    }
    //return (String[]) v.toArray(STRING_ARRAY_0); // "toArray" is not available in JDK 1.1
    String[] a = new String[v.size()];
    for (int i = 0; i < a.length; i++) {
      a[i] = (String) v.elementAt(i);
    }
    return a;
  }

  private String doCommand(String s) throws IOException {
    out.println(s);
    return readReply();
  }

  private String readReply() throws IOException {
    String r = in.readLine();
    if ((r == null) || (r.length() == 0) || (r.charAt(0) != '+')) {
      if (POP3Applet.DEBUG) {
        System.out.println("DUMP negative response: " + r);
      }
      throw new POP3Exception();
    }
    return r;
  }
}
