package nick.inetjava.server;

import java.net.Socket;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.io.*;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Processes a single 'get' request and dies.
 *
 * @author  Nick Nickolov, nick_nickolov@abv.bg
 * @version 16-Apr-02, 09:48:25
 */

final class ProcessingThread extends Thread {
  private SimpleWebServer owner;
  private Socket socket;
  private byte[] tempBuffer = new byte[1024];
  private VirtualFileSystem vfs;

  /**
   * A base url, in case the name of the resource is only a filename
   */
  private static final URL URL0;
  static {
    try {
      URL0 = new URL("http://localhost");
    } catch (MalformedURLException mue) {
      throw new RuntimeException("Can't create a loopback URL");
    }
  }

  protected ProcessingThread(SimpleWebServer owner, Socket socket) {
    this.owner = owner;
    this.socket = socket;
    this.vfs = owner.getVirtualFileSystem();
  }

  public void run() {
    try {
      InputStream in0 = socket.getInputStream();
      OutputStream out0 = socket.getOutputStream();
      BufferedReader in = new BufferedReader(new InputStreamReader(in0));
      PrintStream out = new PrintStream(out0);
      // Parse the request
      String s = in.readLine();
      owner.log("First line of request: " + s);
      if (s == null) {
        makeResponseErrorNotFound(out);
        return;
      }
      while (true) {
        String temp = in.readLine();
        if ((temp == null) || (temp.length() == 0)) {
          break;
        }
      }
      if (!s.startsWith("GET ")) {
        makeResponseErrorNotImplemented(out);
        return;
      }
      s = s.substring(4).trim();
      int indexOfSpace = s.indexOf(' ');
      s = (indexOfSpace == -1) ? s : s.substring(0, indexOfSpace);
      s = new URL(URL0, s).getFile();
      s = unescape(s);
      owner.log("Name of resource: " + s);
      // Resolve the resource name to a physical name
      VirtualFile vf = vfs.get(s);
      if (vf == null) {
        makeResponseErrorNotFound(out);
        return;
      }
      owner.log("Physical location: " + vf.getFile());
      if (vf.isDirectory()) {
        makeResponseDir(out, vf, s);
        return;
      } else {
        makeResponseFile(out, out0, vf.getFile());
        return;
      }
    } catch (IOException ioe) {
      owner.onError("", ioe);
    }
  }

  private void makeResponseErrorNotFound(PrintStream out) throws IOException {
    owner.log("makeResponseErrorNotFound");
    out.println("HTTP/1.0 404 Not Found");
    out.println("MIME-Version: 1.0");
    out.println("Content-type: text/html");
    out.println();
    out.println("<html> Requested document not found! </html>");
    out.println();
    socket.close();
  }

  private void makeResponseErrorNotImplemented(PrintStream out) throws IOException {
    owner.log("makeResponseErrorNotImplemented");
    out.println("HTTP/1.0 501 Not Implemented");
    out.println("MIME-Version: 1.0");
    out.println("Content-type: text/html");
    out.println();
    out.println("<html> Not implemented </html>");
    socket.close();
  }

  private void makeResponseFile(PrintStream out, OutputStream out0, File file) throws IOException {
    owner.log("makeResponseFile");
    out.println("HTTP/1.0 200 OK");
    out.println("MIME-Version: 1.0");
    String filename = file.getName();
    int indexOfDot = filename.indexOf('.');
    String ext = (indexOfDot == -1) ? "" : filename.substring(indexOfDot + 1);
    String ct = owner.getProperty("mime." + ext);
    if (ct == null) {
      ct = owner.getProperty("mime");
      if (ct == null) {
        ct = "text/plain";
      }
    }
    owner.log("Content-type: " + ct);
    out.println("Content-type: " + ct);
    out.println();
    out.flush();
    FileInputStream fin = new FileInputStream(file);
    try {
      while (true) {
        int x = fin.read(tempBuffer);
        if (x == -1) {
          break;
        }
        out0.write(tempBuffer, 0, x);
      }
      fin.close();
    } catch (SocketException se) {
      owner.log("SocketException while writing the response: " + se);
    }
    socket.close();
  }

  private void makeResponseDir(PrintStream out, VirtualFile dir, String resourceName) throws IOException {
    out.println("HTTP/1.0 200 OK");
    out.println("MIME-Version: 1.0");
    out.println("Content-type: text/html");
    out.println();
    out.println("<html><table border=1 width=100%>");
    out.println("<caption><h1>" + resourceName + "</h1></caption>");
    if (dir.getPath().length != 0) {
      out.println("<tr><td><a href='..'>[..]</a><td>&nbsp;");
    }
    String[] children = dir.list();
    File dirPhysical = dir.getFile();
    File[] files = (dirPhysical == null) ? null : dirPhysical.listFiles();
    if (children != null) {
      for (int i = 0; i < children.length; i++) {
        //out.print("<tr><td><a href='" + baseURL + escape(children[i]) + "'>" + children[i] + "</a><td>");
        out.print("<tr><td><a href='" + dir + escape(children[i]) + "'>" + children[i] + "</a><td>");
        out.println(((files == null) || files[i].isDirectory()) ? "&nbsp" : formatFileSize(files[i]));
      }
    }
    out.println("</table><div align=right><i>Simple Web Server, &copy; Nick Nickolov 2002</i></div></html>");
    socket.close();
  }

  /**
   * <tt>descendant</tt> should be a canonical file
   */
  private boolean isAncestor(File ancestor, File descendant) throws IOException {
    while (descendant != null) {
      if (ancestor.equals(descendant)) {
        return true;
      }
      descendant = descendant.getParentFile();
    }
    return false;
  }

  private String formatFileSize(File f) {
    long x = f.length();
    if (x < 1024) {
      return x + "B ";
    } else if (x < 1024*1024) {
      return (x / 1024) + "kB";
    } else {
      return (x / (1024*1024)) + "MB";
    }
  }

  private static final char[] INT_TO_HEX = new char[16];
  private static final int[] HEX_TO_INT = new int[256];
  static {
    java.util.Arrays.fill(HEX_TO_INT, -1);
    for (int i = 0; i < 10; i++) {
      INT_TO_HEX[i] = (char) ('0' + i);
      HEX_TO_INT[(char) ('0' + i)] = i;
    }
    for (int i = 0; i < 6; i++) {
      INT_TO_HEX[10 + i] = (char) ('A' + i);
      HEX_TO_INT[(char) ('A' + i)] = 10 + i;
      HEX_TO_INT[(char) ('a' + i)] = 10 + i;
    }
  }

  private String escape(String s) {
    int ls = s.length();
    StringBuffer b = new StringBuffer();
    for (int i = 0; i < ls; i++) {
      char ch = s.charAt(i);
      if (((ch >= '0') && (ch <= '9')) ||
          ((ch >= 'a') && (ch <= 'z')) ||
          ((ch >= 'A') && (ch <= 'Z')) ||
          (ch == '+') || (ch == '-') || (ch == '*') ||
          (ch == '/') || (ch == '.') || (ch == '_') || (ch == '@')) {
        b.append(ch);
      } else if (ch <= 0xFF) {
        b.append('%').append(INT_TO_HEX[ch >> 4]).append(INT_TO_HEX[ch & 0xF]);
      } else {
        b.append("%u").append(INT_TO_HEX[(ch >> 12) & 0xF])
                      .append(INT_TO_HEX[(ch >> 8) & 0xF])
                      .append(INT_TO_HEX[(ch >> 4) & 0xF])
                      .append(INT_TO_HEX[ch & 0xF]);
      }
    }
    return b.toString();
  }

  private String unescape(String s) {
    StringBuffer b = new StringBuffer();
    int ls = s.length();
    LABEL_1:
    for (int i = 0; i < ls; i++) {
      char ch = s.charAt(i);
      LABEL_2:
      if (ch == '%') {
        try {
          char ch1 = s.charAt(i + 1);
          char ch2 = s.charAt(i + 2);
          if (ch1 == '%') {
            char ch3 = s.charAt(i + 3);
            char ch4 = s.charAt(i + 4);
            char ch5 = s.charAt(i + 5);
            int x = hexToDec(ch2, ch3, ch4, ch5);
            if (x == -1) {
              break LABEL_2;
            }
            b.append((char) x);
            i += 4;
          } else {
            int x = hexToDec(ch1, ch2);
            if (x == -1) {
              break LABEL_2;
            }
            b.append((char) x);
            i += 2;
          }
          continue LABEL_1;
        } catch (StringIndexOutOfBoundsException e) {
          // Invalid escaping - ignore
        }
      }
      b.append(ch);
    }
    owner.log("unescape('" + s + "') -> '" + b + "'");
    return b.toString();
  }

  private int hexToDec(char c0, char c1, char c2, char c3) {
    int x = hexToDec(c0, c1);
    int y = hexToDec(c2, c3);
    return ((x == -1) || (y == -1)) ? -1 : ((x << 8) | y);
  }

  private int hexToDec(char cx, char cy) {
    int x = hexToDec(cx);
    int y = hexToDec(cy);
    return ((x == -1) || (y == -1)) ? -1 : ((x << 4) | y);
  }

  private int hexToDec(char ch) {
    if ((ch >= '0') && (ch <= '9')) {
      return (ch - '0');
    }
    if ((ch >= 'a') && (ch <= 'h')) {
      return (ch - 'a' + 10);
    }
    if ((ch >= 'A') && (ch <= 'H')) {
      return (ch - 'A' + 10);
    }
    return -1;
  }
}
