package nick.inetjava.server;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * @author  Nick Nickolov, nick_nickolov@abv.bg
 * @version 16-Apr-02, 09:40:40
 */

public final class SimpleWebServer extends Thread {
  private boolean debug = false;
  private Properties conf;
  private VirtualFileSystem vfs;
  private int port;

  public SimpleWebServer(Properties conf) throws IOException {
    this.conf = conf;
    this.debug = conf.getProperty("debug", "0").equals("1");
    this.port = -1;
    String s = conf.getProperty("port");
    if (s != null) {
      try {
        this.port = Integer.parseInt(s);
      } catch (NumberFormatException nfe) {
      }
    }
    this.port = (this.port < 0) ? 80 : this.port;
    this.vfs = new VirtualFileSystem(this);
  }

  public void run() {
    ServerSocket serverSocket;
    try {
      serverSocket = new ServerSocket(port);
    } catch (BindException be) {
      System.out.println("Server can't listen at port :" + port + " because it has already been bound");
      return;
    } catch (IOException ioe) {
      onError("", ioe);
      return;
    }
    while (true) {
      Socket socket;
      try {
        socket = serverSocket.accept();
        log("Socket accepted");
        ProcessingThread pth = new ProcessingThread(this, socket);
        pth.start();
      } catch (IOException ioe) {
        onError("", ioe);
      }
    }
  }

  protected void onError(String s, Exception e) {
    if (!debug) {
      return;
    }
    System.out.println("Error: " + s);
    if (e != null) {
      e.printStackTrace();
    }
  }

  protected void onWarning(String s, Exception e) {
    if (!debug) {
      return;
    }
    System.out.println("Warning: " + s);
    if (e != null) {
      e.printStackTrace();
    }
  }

  protected void log(String s) {
    if (!debug) {
      return;
    }
    System.out.println("Log: " + s);
  }

  protected String getProperty(String key) {
    return conf.getProperty(key);
  }

  protected Properties getConf() {
    return conf;
  }

  protected VirtualFileSystem getVirtualFileSystem() {
    return vfs;
  }

  private static void printUsage() {
    System.out.println();
    System.out.println("Usage:");
    System.out.println("        server [-debug] [-conf <conf_file>] <root_dir>");
  }

  public static void main(String[] args) throws Exception {
    System.out.println();
    System.out.println("Simple Web Server, InetJava 2002, FMI");
    System.out.println("Nick Nickolov, fn 42716");
    System.out.println();
    String rootFile = null;
    boolean optionDebug = false;
    String confFilename = null;
    boolean optionHelp = false;
    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-debug")) {
        optionDebug = true;
      } else if (args[i].equals("-conf") && (i < args.length - 1)) {
        confFilename = args[i + 1];
      } else if (args[i].equals("-help") || args[i].equals("-?")) {
        optionHelp = true;
      } else {
        rootFile = args[i];
      }
    }
    if (optionHelp) {
      printUsage();
      return;
    }
    if (confFilename == null) {
      confFilename = "conf.properties";
    }
    Properties conf = new Properties();
    try {
      conf.load(new FileInputStream(confFilename));
    } catch (IOException ioe) {
      try {
        conf.load(SimpleWebServer.class.getResourceAsStream("conf.properties"));
      } catch (IOException ioe1) {
        System.out.println("Can't find the conf file, server not started");
        return;
      }
    }
    if (optionDebug) {
      conf.setProperty("debug", "1");
    }
    if (rootFile != null) {
      conf.setProperty("root", rootFile);
    }
    System.out.println("Server started");
    SimpleWebServer server = new SimpleWebServer(conf);
    server.start();
  }
}
